@@ -1,5 +1,5 @@
 /* 
-Copyright (C) 1993 Free Software Foundation
+Copyright (C) 1993, 1995 Free Software Foundation
 
 This file is part of the GNU IO Library.  This library is free
 software; you can redistribute it and/or modify it under the
@@ -13,8 +13,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+along with this library; see the file COPYING.  If not, write to the Free
+Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 As a special exception, if you link this library with files
 compiled with a GNU compiler to produce an executable, this does not cause
@@ -43,21 +43,26 @@ extern int errno;
    is that of gptr(); in put mode that of pptr().
 
    The position in the buffer that corresponds to the position
-   in external file system is file_ptr().
-   This is normally _IO_read_end, except in putback mode,
-   when it is _IO_save_end.
+   in external file system is normally _IO_read_end, except in putback
+   mode, when it is _IO_save_end.
    If the field _fb._offset is >= 0, it gives the offset in
-   the file as a whole corresponding to eGptr(). (???)
+   the file as a whole corresponding to eGptr(). (?)
 
    PUT MODE:
-   If a filebuf is in put mode, pbase() is non-NULL and equal to base().
-   Also, epptr() == ebuf().
-   Also, eback() == gptr() && gptr() == egptr().
-   The un-flushed character are those between pbase() and pptr().
+   If a filebuf is in put mode, then all of _IO_read_ptr, _IO_read_end,
+   and _IO_read_base are equal to each other.  These are usually equal
+   to _IO_buf_base, though not necessarily if we have switched from
+   get mode to put mode.  (The reason is to maintain the invariant
+   that _IO_read_end corresponds to the external file position.)
+   _IO_write_base is non-NULL and usually equal to _IO_base_base.
+   We also have _IO_write_end == _IO_buf_end, but only in fully buffered mode.
+   The un-flushed character are those between _IO_write_base and _IO_write_ptr.
+
    GET MODE:
    If a filebuf is in get or putback mode, eback() != egptr().
    In get mode, the unread characters are between gptr() and egptr().
    The OS file position corresponds to that of egptr().
+
    PUTBACK MODE:
    Putback mode is used to remember "excess" characters that have
    been sputbackc'd in a separate putback buffer.
@@ -70,7 +75,7 @@ extern int errno;
    The OS position corresponds to that of save_egptr().
    
    LINE BUFFERED OUTPUT:
-   During line buffered output, pbase()==base() && epptr()==base().
+   During line buffered output, _IO_write_base==base() && epptr()==base().
    However, ptr() may be anywhere between base() and ebuf().
    This forces a call to filebuf::overflow(int C) on every put.
    If there is more space in the buffer, and C is not a '\n',
@@ -85,8 +90,8 @@ extern int errno;
 
 
 void
-_IO_file_init(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_file_init, (fp),
+      register _IO_FILE *fp)
 {
   /* POSIX.1 allows another file handle to be used to change the position
      of our file descriptor.  Hence we actually don't know the actual
@@ -95,22 +100,24 @@ _IO_file_init(fp)
   fp->_IO_file_flags |= CLOSED_FILEBUF_FLAGS;
 
   _IO_link_in(fp);
+#if 0
   fp->_fileno = -1;
+#endif
 }
 
 int
-_IO_file_close_it(fp)
-     register _IO_FILE* fp;
+DEFUN(_IO_file_close_it, (fp),
+      register _IO_FILE* fp)
 {
-  int sync_status, close_status;
+  int write_status, close_status;
   if (!_IO_file_is_open(fp))
     return EOF;
 
-  sync_status = _IO_file_sync (fp);
+  write_status = _IO_do_flush (fp);
 
   _IO_unsave_markers(fp);
 
-  close_status = fp->_jumps->__close(fp);
+  close_status = _IO_SYSCLOSE (fp);
 
   /* Free buffer. */
   _IO_setb(fp, NULL, NULL, 0);
@@ -122,27 +129,25 @@ _IO_file_close_it(fp)
   fp->_fileno = EOF;
   fp->_offset = _IO_pos_BAD;
 
-  return close_status ? close_status : sync_status;
+  return close_status ? close_status : write_status;
 }
 
 void
-_IO_file_finish(fp)
-     register _IO_FILE* fp;
+DEFUN(_IO_file_finish, (fp),
+      register _IO_FILE* fp)
 {
   if (_IO_file_is_open(fp))
     {
       _IO_do_flush (fp);
       if (!(fp->_flags & _IO_DELETE_DONT_CLOSE))
-	fp->_jumps->__close(fp);
+	_IO_SYSCLOSE (fp);
     }
   _IO_default_finish(fp);
 }
 
 _IO_FILE *
-_IO_file_fopen(fp, filename, mode)
-     register _IO_FILE *fp;
-     const char *filename;
-     const char *mode;
+DEFUN(_IO_file_fopen, (fp, filename, mode),
+      register _IO_FILE *fp AND const char *filename AND const char *mode)
 {
   int oflags = 0, omode;
   int read_write, fdesc;
@@ -178,17 +183,16 @@ _IO_file_fopen(fp, filename, mode)
   fp->_fileno = fdesc;
   _IO_mask_flags(fp, read_write,_IO_NO_READS+_IO_NO_WRITES+_IO_IS_APPENDING);
   if (read_write & _IO_IS_APPENDING)
-    if (fp->_jumps->__seekoff(fp, (_IO_off_t)0, _IO_seek_end) == _IO_pos_BAD
-	&& errno != ESPIPE)
+    if (_IO_SEEKOFF (fp, (_IO_off_t)0, _IO_seek_end, _IOS_INPUT|_IOS_OUTPUT)
+      == _IO_pos_BAD && errno != ESPIPE)
       return NULL;
   _IO_link_in(fp);
   return fp;
 }
 
 _IO_FILE*
-_IO_file_attach(fp, fd)
-     _IO_FILE *fp;
-     int fd;
+DEFUN(_IO_file_attach, (fp, fd),
+      _IO_FILE *fp AND int fd)
 {
   if (_IO_file_is_open(fp))
     return NULL;
@@ -198,33 +202,32 @@ _IO_file_attach(fp, fd)
   /* Get the current position of the file. */
   /* We have to do that since that may be junk. */
   fp->_offset = _IO_pos_BAD;
-  if (fp->_jumps->__seekoff(fp, (_IO_off_t) 0, _IO_seek_cur)
+  if (_IO_SEEKOFF (fp, (_IO_off_t)0, _IO_seek_cur, _IOS_INPUT|_IOS_OUTPUT)
       == _IO_pos_BAD && errno != ESPIPE)
     return NULL;
   return fp;
 }
 
-int
-_IO_file_setbuf(fp, p, len)
-     register _IO_FILE *fp;
-     char* p;
-     _IO_ssize_t len;
+_IO_FILE*
+DEFUN(_IO_file_setbuf, (fp, p, len),
+      register _IO_FILE *fp AND char* p AND _IO_ssize_t len)
 {
-    if (_IO_default_setbuf(fp, p, len) == EOF)
-	return EOF;
+    if (_IO_default_setbuf(fp, p, len) == NULL)
+	return NULL;
 
     fp->_IO_write_base = fp->_IO_write_ptr = fp->_IO_write_end
       = fp->_IO_buf_base;
     _IO_setg(fp, fp->_IO_buf_base, fp->_IO_buf_base, fp->_IO_buf_base);
 
-    return 0;
+    return fp;
 }
 
+/* Write TO_DO bytes from DATA to FP.
+   Then mark FP has having empty buffers. */
+
 int
-_IO_do_write(fp, data, to_do)
-     register _IO_FILE *fp;
-     const char* data;
-     _IO_size_t to_do;
+DEFUN(_IO_do_write, (fp, data, to_do),
+  register _IO_FILE *fp AND const char* data AND _IO_size_t to_do)
 {
   _IO_size_t count;
   if (to_do == 0)
@@ -238,12 +241,13 @@ _IO_do_write(fp, data, to_do)
     fp->_offset = _IO_pos_BAD;
   else if (fp->_IO_read_end != fp->_IO_write_base)
     { 
-      _IO_pos_t new_pos = fp->_jumps->__seek(fp, fp->_IO_write_base - fp->_IO_read_end, 1);
+      _IO_pos_t new_pos
+	= _IO_SYSSEEK(fp, fp->_IO_write_base - fp->_IO_read_end, 1);
       if (new_pos == _IO_pos_BAD)
 	return EOF;
       fp->_offset = new_pos;
     }
-  count = fp->_jumps->__write(fp, data, to_do);
+  count = _IO_SYSWRITE (fp, data, to_do);
   if (fp->_cur_column)
     fp->_cur_column = _IO_adjust_column(fp->_cur_column - 1, data, to_do) + 1;
   _IO_setg(fp, fp->_IO_buf_base, fp->_IO_buf_base, fp->_IO_buf_base);
@@ -254,8 +258,8 @@ _IO_do_write(fp, data, to_do)
 }
 
 int
-_IO_file_underflow(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_file_underflow, (fp),
+      register _IO_FILE *fp)
 {
   _IO_ssize_t count;
 #if 0
@@ -279,8 +283,8 @@ _IO_file_underflow(fp)
 
   _IO_switch_to_get_mode(fp);
 
-  count = fp->_jumps->__read(fp, fp->_IO_buf_base,
-			     fp->_IO_buf_end - fp->_IO_buf_base);
+  count = _IO_SYSREAD (fp, fp->_IO_buf_base,
+		       fp->_IO_buf_end - fp->_IO_buf_base);
   if (count <= 0)
     {
       /* 
@@ -303,7 +307,7 @@ _IO_file_underflow(fp)
 #endif
 	fp->_flags |= _IO_EOF_SEEN;
       else
-        fp->_flags |= _IO_ERR_SEEN;
+	fp->_flags |= _IO_ERR_SEEN;
       count = 0;
   }
   fp->_IO_read_base = fp->_IO_read_ptr = fp->_IO_buf_base;
@@ -317,9 +321,9 @@ _IO_file_underflow(fp)
   return *(unsigned char*)fp->_IO_read_ptr;
 }
 
-int _IO_file_overflow (f, ch)
-     register _IO_FILE* f;
-     int ch;
+int
+DEFUN(_IO_file_overflow, (f, ch),
+      register _IO_FILE* f AND int ch)
 {
   if (f->_flags & _IO_NO_WRITES) /* SET ERROR */
     return EOF;
@@ -332,17 +336,17 @@ int _IO_file_overflow (f, ch)
 	  _IO_doallocbuf(f);
 	  _IO_setg (f, f->_IO_buf_base, f->_IO_buf_base, f->_IO_buf_base);
 	}
-      /* Otherwise must be currently reading. */
+      /* Otherwise must be currently reading.
+	 If _IO_read_ptr (and hence also _IO_read_end) is at the buffer end,
+	 logically slide the buffer forwards one block (by setting the
+	 read pointers to all point at the beginning of the block).  This
+	 makes room for subsequent output.
+	 Otherwise, set the read pointers to _IO_read_end (leaving that
+	 alone, so it can continue to correspond to the external position). */
       if (f->_IO_read_ptr == f->_IO_buf_end)
-      {
-        f->_IO_write_base = f->_IO_write_ptr;
-        f->_IO_write_ptr = f->_IO_read_ptr;
-      }
-      else
-      {
-        f->_IO_write_ptr = f->_IO_read_ptr;
-        f->_IO_write_base = f->_IO_write_ptr;
-      }
+	f->_IO_read_end = f->_IO_read_ptr = f->_IO_buf_base;
+      f->_IO_write_ptr = f->_IO_read_ptr;
+      f->_IO_write_base = f->_IO_write_ptr;
       f->_IO_write_end = f->_IO_buf_end;
       f->_IO_read_base = f->_IO_read_ptr = f->_IO_read_end;
 
@@ -364,8 +368,8 @@ int _IO_file_overflow (f, ch)
 }
 
 int
-_IO_file_sync(fp)
-     register _IO_FILE* fp;
+DEFUN(_IO_file_sync, (fp),
+      register _IO_FILE* fp)
 {
   _IO_size_t delta;
   /*    char* ptr = cur_ptr(); */
@@ -378,7 +382,7 @@ _IO_file_sync(fp)
       if (_IO_in_backup(fp))
 	delta -= eGptr() - Gbase();
 #endif
-      _IO_off_t new_pos = fp->_jumps->__seek(fp, delta, 1);
+      _IO_off_t new_pos = _IO_SYSSEEK (fp, delta, 1);
       if (new_pos != (_IO_off_t)EOF)
 	fp->_IO_read_end = fp->_IO_read_ptr;
 #ifdef ESPIPE
@@ -395,17 +399,14 @@ _IO_file_sync(fp)
 }
 
 _IO_pos_t
-_IO_file_seekoff(fp, offset, mode)
-     register _IO_FILE *fp;
-     _IO_off_t offset;
-     _IO_seekflags mode;
+DEFUN(_IO_file_seekoff, (fp, offset, dir, mode),
+      register _IO_FILE *fp AND _IO_off_t offset AND int dir AND int mode)
 {
   _IO_pos_t result;
   _IO_off_t delta, new_offset;
   long count;
-  int dir = mode & 3;
 
-  if ((mode & _IO_seek_not_in) && (mode & _IO_seek_not_out))
+  if (mode == 0)
     dir = _IO_seek_cur, offset = 0; /* Don't move any pointers. */
 
   /* Flush unwritten characters.
@@ -443,7 +444,7 @@ _IO_file_seekoff(fp, offset, mode)
     case _IO_seek_end:
       {
 	struct stat st;
-	if (fp->_jumps->__stat(fp, &st) == 0 && S_ISREG(st.st_mode))
+	if (_IO_SYSSTAT (fp, &st) == 0 && S_ISREG(st.st_mode))
 	  {
 	    offset += st.st_size;
 	    dir = _IO_seek_set;
@@ -455,15 +456,8 @@ _IO_file_seekoff(fp, offset, mode)
   /* At this point, dir==_IO_seek_set. */
 
   /* If destination is within current buffer, optimize: */
-#if 0
-  /* Don't do it if it is opened for write. Some other processes
-   * may be writing into it. */
-  if ((fp->_flags & _IO_NO_WRITES) && fp->_offset != _IO_pos_BAD \
-	&& fp->_IO_read_base != NULL && !_IO_in_backup (fp))
-#else
   if (fp->_offset != _IO_pos_BAD && fp->_IO_read_base != NULL
       && !_IO_in_backup (fp))
-#endif
     {
       /* Offset relative to start of main get area. */
       _IO_pos_t rel_offset = offset - fp->_offset
@@ -519,15 +513,15 @@ _IO_file_seekoff(fp, offset, mode)
       new_offset = offset;
       delta = 0;
     }
-  result = fp->_jumps->__seek(fp, new_offset, 0);
+  result = _IO_SYSSEEK (fp, new_offset, 0);
   if (result < 0)
     return EOF;
   if (delta == 0)
     count = 0;
   else
     {
-      count = fp->_jumps->__read(fp, fp->_IO_buf_base,
-				 fp->_IO_buf_end - fp->_IO_buf_base);
+      count = _IO_SYSREAD (fp, fp->_IO_buf_base,
+			   fp->_IO_buf_end - fp->_IO_buf_base);
       if (count < delta)
 	{
 	  /* We weren't allowed to read, but try to seek the remainder. */
@@ -544,7 +538,7 @@ _IO_file_seekoff(fp, offset, mode)
  dumb:
 
   _IO_unsave_markers(fp);
-  result = fp->_jumps->__seek(fp, offset, dir);
+  result = _IO_SYSSEEK (fp, offset, dir);
   if (result != EOF)
     _IO_mask_flags(fp, 0, _IO_EOF_SEEN);
   fp->_offset = result;
@@ -554,10 +548,8 @@ _IO_file_seekoff(fp, offset, mode)
 }
 
 _IO_ssize_t
-_IO_file_read(fp, buf, size)
-     register _IO_FILE* fp;
-     void* buf;
-     _IO_ssize_t size;
+DEFUN(_IO_file_read, (fp, buf, size),
+      register _IO_FILE* fp AND void* buf AND _IO_ssize_t size)
 {
   for (;;)
     {
@@ -571,34 +563,29 @@ _IO_file_read(fp, buf, size)
 }
 
 _IO_pos_t
-_IO_file_seek(fp, offset, dir)
-     _IO_FILE *fp;
-     _IO_off_t offset;
-     int dir;
+DEFUN(_IO_file_seek, (fp, offset, dir),
+      _IO_FILE *fp AND _IO_off_t offset AND int dir)
 {
   return _IO_lseek(fp->_fileno, offset, dir);
 }
 
 int
-_IO_file_stat(fp, st)
-     _IO_FILE *fp;
-     void* st;
+DEFUN(_IO_file_stat, (fp, st),
+      _IO_FILE *fp AND void* st)
 {
   return _IO_fstat(fp->_fileno, (struct stat*)st);
 }
 
 int
-_IO_file_close(fp)
-     _IO_FILE* fp;
+DEFUN(_IO_file_close, (fp),
+      _IO_FILE* fp)
 {
   return _IO_close(fp->_fileno);
 }
 
 _IO_ssize_t
-_IO_file_write(f, data, n)
-     register _IO_FILE* f;
-     const void* data;
-     _IO_ssize_t n;
+DEFUN(_IO_file_write, (f, data, n),
+      register _IO_FILE* f AND const void* data AND _IO_ssize_t n)
 {
   _IO_ssize_t to_do = n;
   while (to_do > 0)
@@ -626,12 +613,10 @@ _IO_file_write(f, data, n)
 }
 
 _IO_size_t
-_IO_file_xsputn(f, data, n)
-     _IO_FILE *f;
-     const void *data;
-     _IO_size_t n;
+DEFUN(_IO_file_xsputn, (f, data, n),
+      _IO_FILE *f AND const void *data AND _IO_size_t n)
 {
-  register const char *s = data;
+  register const char *s = (char*) data;
   _IO_size_t to_do = n;
   int must_flush = 0;
   _IO_size_t count;
@@ -705,10 +690,8 @@ _IO_file_xsputn(f, data, n)
 #if 0
 /* Work in progress */
 _IO_size_t
-_IO_file_xsgetn(fp, data, n)
-     _IO_FILE *fp;
-     void *data;
-     _IO_size_t n;
+DEFUN(_IO_file_xsgetn, (fp, data, n),
+      _IO_FILE *fp AND void *data AND _IO_size_t n)
 {
   register _IO_size_t more = n;
   register char *s = data;
@@ -751,7 +734,7 @@ _IO_file_xsgetn(fp, data, n)
 	    _IO_flush_all_linebuffered();
 
 	  _IO_switch_to_get_mode(fp); ???;
-	  count = fp->_jumps->__read (fp, s, more);
+	  count = _IO_SYSREAD (fp, s, more);
 	  if (count <= 0)
 	     {
 	       if (count == 0)
@@ -772,21 +755,22 @@ _IO_file_xsgetn(fp, data, n)
 #endif
 
 struct _IO_jump_t _IO_file_jumps = {
-  _IO_file_overflow,
-  _IO_file_underflow,
-  _IO_file_xsputn,
-  _IO_default_xsgetn,
-  _IO_file_read,
-  _IO_file_write,
-  _IO_file_doallocate,
-  _IO_default_pbackfail,
-  _IO_file_setbuf,
-  _IO_file_sync,
-  _IO_file_finish,
-  _IO_file_close,
-  _IO_file_stat,
-  _IO_file_seek,
-  _IO_file_seekoff,
-  _IO_default_seekpos,
-  _IO_default_uflow
+  JUMP_INIT_DUMMY,
+  JUMP_INIT(finish, _IO_file_finish),
+  JUMP_INIT(overflow, _IO_file_overflow),
+  JUMP_INIT(underflow, _IO_file_underflow),
+  JUMP_INIT(uflow, _IO_default_uflow),
+  JUMP_INIT(pbackfail, _IO_default_pbackfail),
+  JUMP_INIT(xsputn, _IO_file_xsputn),
+  JUMP_INIT(xsgetn, _IO_default_xsgetn),
+  JUMP_INIT(seekoff, _IO_file_seekoff),
+  JUMP_INIT(seekpos, _IO_default_seekpos),
+  JUMP_INIT(setbuf, _IO_file_setbuf),
+  JUMP_INIT(sync, _IO_file_sync),
+  JUMP_INIT(doallocate, _IO_file_doallocate),
+  JUMP_INIT(read, _IO_file_read),
+  JUMP_INIT(write, _IO_file_write),
+  JUMP_INIT(seek, _IO_file_seek),
+  JUMP_INIT(close, _IO_file_close),
+  JUMP_INIT(stat, _IO_file_stat)
 };
