@@ -126,12 +126,14 @@ static unsigned int guess_grouping (unsigned int intdig_max,
 static char *group_number (char *buf, char *bufend, unsigned int intdig_no,
 			   const char *grouping, wchar_t thousands_sep);
 
+int __printf_fp( FILE *, const struct printf_info *, const ** );
+static unsigned int guess_grouping( unsigned int, const char *, wchar_t );
+static char *group_number( char *, char *, unsigned int,
+								  const char *, wchar_t );
+char hack_digit __P (( void ));
 
 int
-__printf_fp (fp, info, args)
-     FILE *fp;
-     const struct printf_info *info;
-     const **const args;
+__printf_fp( FILE *fp, const struct printf_info *info, const **const args )
 {
   /* The floating-point value to output.  */
   union
@@ -157,7 +159,7 @@ __printf_fp (fp, info, args)
   /* We need to shift the contents of fp_input by this amount of bits.	*/
   int to_shift;
 
-  /* The significant of the floting-point value in question  */
+  /* The significant of the floating-point value in question  */
   MPN_VAR(frac);
   /* and the exponent.	*/
   int exponent;
@@ -184,6 +186,9 @@ __printf_fp (fp, info, args)
   /* General helper (carry limb).  */
   mp_limb cy;
 
+  /* This is supposed to be the way to prototype a nested function. vch */
+  auto char hack_digit (void);
+
   char hack_digit (void)
     {
       mp_limb hi;
@@ -222,7 +227,7 @@ __printf_fp (fp, info, args)
 	}
 
       return '0' + hi;
-    }
+    }	/* end of 'hack_digit()' */
 
 
   /* Figure out the decimal point character.  */
@@ -234,7 +239,11 @@ __printf_fp (fp, info, args)
   if (info->group)
     {
       grouping = _NL_CURRENT (LC_NUMERIC, GROUPING);
-      if (*grouping <= 0 || *grouping == CHAR_MAX)
+#ifdef __CHAR_UNSIGNED__
+      if (*grouping == CHAR_MAX )
+#else
+      if (*grouping == CHAR_MAX || *grouping < 0)
+#endif
 	grouping = NULL;
       else
 	{
@@ -249,6 +258,7 @@ __printf_fp (fp, info, args)
   else
     grouping = NULL;
 
+  to_shift = 0;
   /* Fetch the argument value.	*/
   if (info->is_long_double && sizeof (long double) > sizeof (double))
     {
@@ -935,7 +945,11 @@ guess_grouping (unsigned int intdig_max, const char *grouping, wchar_t sepchar)
 
   /* We treat all negative values like CHAR_MAX.  */
 
-  if (*grouping == CHAR_MAX || *grouping <= 0)
+#ifdef __CHAR_UNSIGNED__
+  if (*grouping == CHAR_MAX )
+#else
+  if (*grouping == CHAR_MAX || *grouping < 0)
+#endif
     /* No grouping should be done.  */
     return 0;
 
@@ -945,13 +959,17 @@ guess_grouping (unsigned int intdig_max, const char *grouping, wchar_t sepchar)
       ++groups;
       intdig_max -= *grouping++;
 
+#ifdef __CHAR_UNSIGNED__
+      if (*grouping == CHAR_MAX )
+#else
       if (*grouping == CHAR_MAX || *grouping < 0)
+#endif
 	/* No more grouping should be done.  */
 	break;
       else if (*grouping == 0)
 	{
 	  /* Same grouping repeats.  */
-	  groups += intdig_max / grouping[-1];
+	  groups += (intdig_max - 1) / grouping[-1];
 	  break;
 	}
     }
@@ -986,7 +1004,11 @@ group_number (char *buf, char *bufend, unsigned int intdig_no,
       while (--len > 0);
       *p-- = thousands_sep;
 
+#ifdef __CHAR_UNSIGNED__
+      if (*grouping == CHAR_MAX )
+#else
       if (*grouping == CHAR_MAX || *grouping < 0)
+#endif
 	/* No more grouping should be done.  */
 	break;
       else if (*grouping == 0)
