@@ -536,10 +536,12 @@ printf ("strcmp 9: %d\n", strcmp("a\203", "a\003"));
   /* strerror - VERY system-dependent.  */
   {
     int f;
+    int saved;
     it = "strerror";
     f = open("/", O_WRONLY);	/* Should always fail. */
     check(f < 0 && errno > 0 && errno < sys_nerr, 1);
-    equal(strerror(errno), sys_errlist[errno], 2);
+    saved = errno;
+    equal(strerror(saved), sys_errlist[saved], 2);
   }
 
   {
